@@ -14,6 +14,7 @@ module Pantry
   , defaultHackageSecurityConfig
   , defaultCasaRepoPrefix
   , defaultCasaMaxPerRequest
+  , defaultSnapshotLocation
   , HasPantryConfig (..)
   , withPantryConfig
   , HpackExecutable (..)
@@ -66,7 +67,10 @@ module Pantry
     -- ** Repos
   , Repo (..)
   , RepoType (..)
+  , SimpleRepo (..)
   , withRepo
+  , fetchRepos
+  , fetchReposRaw
 
     -- ** Package location
   , RawPackageLocation (..)
@@ -87,6 +91,8 @@ module Pantry
   , SnapshotLayer (..)
   , toRawSnapshotLayer
   , WantedCompiler (..)
+  , SnapName (..)
+  , snapshotLocation
 
     -- * Loading values
   , resolvePaths
@@ -111,6 +117,7 @@ module Pantry
 
     -- * Parsers
   , parseWantedCompiler
+  , parseSnapName
   , parseRawSnapshotLocation
   , parsePackageIdentifierRevision
   , parseHackageText
@@ -123,10 +130,6 @@ module Pantry
   , parseVersion
   , parseVersionThrowing
 
-    -- * Stackage snapshots
-  , ltsSnapshotLocation
-  , nightlySnapshotLocation
-
     -- * Cabal helpers
   , packageIdentifierString
   , packageNameString
@@ -193,7 +196,7 @@ import Pantry.Casa
 import Casa.Client (thParserCasaRepo, CasaRepoPrefix)
 import Pantry.Repo
 import qualified Pantry.SHA256 as SHA256
-import Pantry.Storage hiding (TreeEntry, PackageName, Version)
+import Pantry.Storage hiding (TreeEntry, PackageName, Version, findOrGenerateCabalFile)
 import Pantry.Tree
 import Pantry.Types as P
 import Pantry.Hackage
@@ -239,10 +242,12 @@ withPantryConfig
   -- ^ The casa pull URL e.g. https://casa.fpcomplete.com/v1/pull.
   -> Int
   -- ^ Max casa keys to pull per request.
+  -> (SnapName -> RawSnapshotLocation)
+  -- ^ The location of snapshot synonyms
   -> (PantryConfig -> RIO env a)
   -- ^ What to do with the config
   -> RIO env a
-withPantryConfig root hsc he count pullURL maxPerRequest inner = do
+withPantryConfig root hsc he count pullURL maxPerRequest snapLoc inner = do
   env <- ask
   pantryRelFile <- parseRelFile "pantry.sqlite3"
   -- Silence persistent's logging output, which is really noisy
@@ -261,6 +266,7 @@ withPantryConfig root hsc he count pullURL maxPerRequest inner = do
       , pcParsedCabalFilesMutable = ref2
       , pcCasaRepoPrefix = pullURL
       , pcCasaMaxPerRequest = maxPerRequest
+      , pcSnapshotLocation = snapLoc
       }
 
 -- | Default pull URL for Casa.
@@ -292,7 +298,7 @@ defaultHackageSecurityConfig = HackageSecurityConfig
       , "fe331502606802feac15e514d9b9ea83fee8b6ffef71335479a2e68d84adc6b0"
       ]
   , hscKeyThreshold = 3
-  , hscDownloadPrefix = "https://s3.amazonaws.com/hackage.fpcomplete.com/"
+  , hscDownloadPrefix = "https://hackage.haskell.org/"
   , hscIgnoreExpiry = False
   }
 
@@ -372,14 +378,15 @@ fetchTreeKeys treeKeys = do
   -- Pull down those tree keys from Casa, automatically inserting into
   -- our local database.
   treeKeyBlobs :: Map TreeKey P.Tree <-
-    fmap
+    handleAny (const mempty)
+    (fmap
       Map.fromList
       (withStorage
          (runConduitRes
             (casaBlobSource
                (fmap unTreeKey (mapMaybe getRawTreeKey packageLocationsMissing)) .|
              mapMC parseTreeM .|
-             sinkList)))
+             sinkList))))
   pullTreeEnd <- liftIO getCurrentTime
   let pulledPackages =
         mapMaybe
@@ -395,21 +402,23 @@ fetchTreeKeys treeKeys = do
           (\(P.TreeMap files) -> Set.fromList (map teBlob (toList files)))
           treeKeyBlobs
   pullBlobStart <- liftIO getCurrentTime
-  pulledBlobKeys :: Int <-
-    withStorage
+  mpulledBlobKeys :: Maybe Int <-
+    handleAny (const (pure Nothing))
+    (fmap Just (withStorage
       (runConduitRes
-         (casaBlobSource uniqueFileBlobKeys .| mapC (const 1) .| sumC))
-  pullBlobEnd <- liftIO getCurrentTime
-  logDebug
-    ("Pulled from Casa: " <>
-     mconcat (List.intersperse ", " (map display pulledPackages)) <>
-     " (" <>
-     display (T.pack (show (diffUTCTime pullTreeEnd pullTreeStart))) <>
-     "), " <>
-     plural pulledBlobKeys "file" <>
-     " (" <>
-     display (T.pack (show (diffUTCTime pullBlobEnd pullBlobStart))) <>
-     ")")
+         (casaBlobSource uniqueFileBlobKeys .| mapC (const 1) .| sumC))))
+  for_ mpulledBlobKeys $ \pulledBlobKeys -> do
+    pullBlobEnd <- liftIO getCurrentTime
+    logDebug
+      ("Pulled from Casa: " <>
+       mconcat (List.intersperse ", " (map display pulledPackages)) <>
+       " (" <>
+       display (T.pack (show (diffUTCTime pullTreeEnd pullTreeStart))) <>
+       "), " <>
+       plural pulledBlobKeys "file" <>
+       " (" <>
+       display (T.pack (show (diffUTCTime pullBlobEnd pullBlobStart))) <>
+       ")")
   -- Store the tree for each missing package.
   for_
     packageLocationsMissing
@@ -432,14 +441,20 @@ fetchTreeKeys treeKeys = do
                    identifier <-
                      getRawPackageLocationIdent rawPackageLocationImmutable
                    case findCabalOrHpackFile rawPackageLocationImmutable tree of
-                     Just buildFile ->
-                       void
-                         (withStorage
-                            (storeTree
-                               rawPackageLocationImmutable
-                               identifier
-                               tree
-                               buildFile))
+                     Just buildFile -> withStorage $ do
+                       ecachedTree <- loadCachedTree tree
+                       case ecachedTree of
+                         Left e ->
+                           lift $ logWarn
+                           ("Loading cached tree after download from Casa failed on " <>
+                            display rawPackageLocationImmutable <> ": " <>
+                            displayShow e)
+                         Right cachedTree ->
+                           void $ storeTree
+                             rawPackageLocationImmutable
+                             identifier
+                             cachedTree
+                             buildFile
                      Nothing ->
                        logWarn
                          ("Unable to find build file for package: " <>
@@ -952,7 +967,7 @@ completePackageLocation pl@(RPLIArchive archive rpm) = do
     Nothing -> byThirdParty (isJust mpackage)
   where
     byThirdParty warnAboutMissingSizeSha = do
-      (sha, size, package) <- getArchive pl archive rpm
+      (sha, size, package, _cachedTree) <- getArchive pl archive rpm
       when warnAboutMissingSizeSha (warnWith sha size)
       -- (getArchive checks archive and package metadata)
       let RawArchive loc _ _ subdir = archive
@@ -1036,6 +1051,8 @@ completeSnapshotLocation (RSLUrl url (Just blobKey)) = pure $ SLUrl url blobKey
 completeSnapshotLocation (RSLUrl url Nothing) = do
   bs <- loadFromURL url Nothing
   pure $ SLUrl url (bsToBlobKey bs)
+completeSnapshotLocation (RSLSynonym syn) =
+  completeSnapshotLocation =<< snapshotLocation syn
 
 traverseConcurrently_
   :: (Foldable f, HasPantryConfig env)
@@ -1424,6 +1441,12 @@ loadRawSnapshotLayer rsl@(RSLFilePath fp) =
     value <- Yaml.decodeFileThrow $ toFilePath $ resolvedAbsolute fp
     snapshot <- warningsParserHelperRaw rsl value $ Just $ parent $ resolvedAbsolute fp
     pure $ Right (snapshot, CompletedSL rsl (SLFilePath fp))
+loadRawSnapshotLayer rsl@(RSLSynonym syn) = do
+  loc <- snapshotLocation syn
+  comp <- loadRawSnapshotLayer loc
+  pure $ case comp of
+    Left wc -> Left wc
+    Right (l, CompletedSL _ n) -> Right (l, CompletedSL rsl n)
 
 -- | Parse a 'SnapshotLayer' value from a 'SnapshotLocation'.
 --
@@ -1665,6 +1688,7 @@ runPantryAppWith maxConnCount casaRepoPrefix casaMaxPerRequest f = runSimpleApp
     maxConnCount
     casaRepoPrefix
     casaMaxPerRequest
+    defaultSnapshotLocation
     $ \pc ->
       runRIO
         PantryApp
@@ -1691,6 +1715,7 @@ runPantryAppClean f = liftIO $ withSystemTempDirectory "pantry-clean" $ \dir ->
     8
     defaultCasaRepoPrefix
     defaultCasaMaxPerRequest
+    defaultSnapshotLocation
     $ \pc ->
       runRIO
         PantryApp
