@@ -15,6 +15,9 @@
 {-# LANGUAGE UndecidableInstances #-}
 {-# LANGUAGE DerivingStrategies #-}
 {-# LANGUAGE StandaloneDeriving #-}
+{-# LANGUAGE TupleSections #-}
+{-# LANGUAGE DataKinds #-}
+{-# LANGUAGE FlexibleInstances #-}
 module Pantry.Storage
   ( SqlBackend
   , initStorage
@@ -65,11 +68,13 @@ module Pantry.Storage
   , getSnapshotCacheId
   , storeSnapshotModuleCache
   , loadExposedModulePackages
+  , findOrGenerateCabalFile
   , PackageNameId
   , PackageName
   , VersionId
   , ModuleNameId
   , Version
+  , versionVersion
   , Unique(..)
   , EntityField(..)
     -- avoid warnings
@@ -90,6 +95,9 @@ module Pantry.Storage
   , UrlBlobId
   , SnapshotCacheId
   , PackageExposedModuleId
+  , loadCachedTree
+  , CachedTree (..)
+  , unCachedTree
   ) where
 
 import RIO hiding (FilePath)
@@ -113,7 +121,7 @@ import Path.IO (listDir, createTempDir, getTempDir, removeDirRecur)
 import Pantry.HPack (hpackVersion, hpack)
 import Conduit
 import Data.Acquire (with)
-import Pantry.Types (PackageNameP (..), VersionP (..), SHA256, FileSize (..), FileType (..), HasPantryConfig, BlobKey, Repo (..), TreeKey, SafeFilePath, Revision (..), Package (..), SnapshotCacheHash (..))
+import Pantry.Types (PackageNameP (..), VersionP (..), SHA256, FileSize (..), FileType (..), HasPantryConfig, BlobKey, Repo (..), TreeKey, SafeFilePath, Revision (..), Package (..), SnapshotCacheHash (..), connRDBMS)
 import qualified Pantry.SQLite as SQLite
 
 share [mkPersist sqlSettings, mkMigrate "migrateAll"] [persistLowerCase|
@@ -264,8 +272,9 @@ withStorage
   :: (HasPantryConfig env, HasLogFunc env)
   => ReaderT SqlBackend (RIO env) a
   -> RIO env a
-withStorage action =
-  flip SQLite.withStorage_ action =<< view (P.pantryConfigL.to P.pcStorage)
+withStorage action = do
+  storage <- view (P.pantryConfigL.to P.pcStorage)
+  SQLite.withStorage_ storage action
 
 -- | This is a helper type to distinguish db queries between different rdbms backends. The important
 -- part is that the affects described in this data type should be semantically equivalent between
@@ -282,7 +291,7 @@ rdbmsAwareQuery
   :: RdbmsActions env a
   -> ReaderT SqlBackend (RIO env) a
 rdbmsAwareQuery RdbmsActions {raSqlite, raPostgres} = do
-  rdbms <- connRDBMS <$> ask
+  rdbms <- Pantry.Types.connRDBMS <$> ask
   case rdbms of
     "postgresql" -> raPostgres
     "sqlite" -> raSqlite
@@ -676,16 +685,23 @@ getFilePathId sfp =
   where
     fp = T.unpack (P.unSafeFilePath sfp)
 
+-- | A tree that has already been stored in the database
+newtype CachedTree
+  = CachedTreeMap (Map SafeFilePath (P.TreeEntry, BlobId))
+  deriving Show
+
+unCachedTree :: CachedTree -> P.Tree
+unCachedTree (CachedTreeMap m) = P.TreeMap $ fst <$> m
 
 storeTree
   :: (HasPantryConfig env, HasLogFunc env, HasProcessContext env)
   => P.RawPackageLocationImmutable -- ^ for exceptions
   -> P.PackageIdentifier
-  -> P.Tree
+  -> CachedTree
   -> P.BuildFile
   -> ReaderT SqlBackend (RIO env) (TreeId, P.TreeKey)
-storeTree rpli (P.PackageIdentifier name version) tree@(P.TreeMap m) buildFile = do
-  (bid, blobKey) <- storeBlob $ P.renderTree tree
+storeTree rpli (P.PackageIdentifier name version) tree@(CachedTreeMap m) buildFile = do
+  (bid, blobKey) <- storeBlob $ P.renderTree $ unCachedTree tree
   (cabalid, ftype) <- case buildFile of
                 P.BFHpack (P.TreeEntry _ ftype) -> pure (Nothing, ftype)
                 P.BFCabal _ (P.TreeEntry (P.BlobKey btypeSha _) ftype) -> do
@@ -708,13 +724,8 @@ storeTree rpli (P.PackageIdentifier name version) tree@(P.TreeMap m) buildFile =
   (tid, pTreeKey) <- case etid of
     Left (Entity tid _) -> pure (tid, P.TreeKey blobKey) -- already in database, assume it matches
     Right tid -> do
-      for_ (Map.toList m) $ \(sfp, P.TreeEntry blobKey' ft) -> do
+      for_ (Map.toList m) $ \(sfp, (P.TreeEntry _blobKey ft, bid')) -> do
         sfpid <- getFilePathId sfp
-        mbid <- getBlobId blobKey'
-        bid' <-
-          case mbid of
-            Nothing -> error $ "Cannot store tree, contains unknown blob: " ++ show blobKey'
-            Just bid' -> pure bid'
         insert_ TreeEntry
           { treeEntryTree = tid
           , treeEntryPath = sfpid
@@ -980,13 +991,12 @@ storeRepoCache repo subdir tid = do
 
 loadRepoCache
   :: Repo
-  -> Text -- ^ subdir
   -> ReaderT SqlBackend (RIO env) (Maybe TreeId)
-loadRepoCache repo subdir = fmap (repoCacheTree . entityVal) <$> selectFirst
+loadRepoCache repo = fmap (repoCacheTree . entityVal) <$> selectFirst
   [ RepoCacheUrl ==. repoUrl repo
   , RepoCacheType ==. repoType repo
   , RepoCacheCommit ==. repoCommit repo
-  , RepoCacheSubdir ==. subdir
+  , RepoCacheSubdir ==. repoSubdir repo
   ]
   [Desc RepoCacheTime]
 
@@ -1174,3 +1184,22 @@ loadExposedModulePackages cacheId mName =
     ]
   where
     go (Single (P.PackageNameP m)) = m
+
+data LoadCachedTreeException = MissingBlob !BlobKey
+  deriving (Show, Typeable)
+instance Exception LoadCachedTreeException
+
+-- | Ensure that all blobs needed for this package are present in the cache
+loadCachedTree :: forall env. P.Tree -> ReaderT SqlBackend (RIO env) (Either LoadCachedTreeException CachedTree)
+loadCachedTree (P.TreeMap m) =
+    try $ CachedTreeMap <$> traverse loadEntry m
+  where
+    loadEntry :: P.TreeEntry -> ReaderT SqlBackend (RIO env) (P.TreeEntry, BlobId)
+    loadEntry te = (te, ) <$> loadBlob' (P.teBlob te)
+
+    loadBlob' :: BlobKey -> ReaderT SqlBackend (RIO env) BlobId
+    loadBlob' blobKey@(P.BlobKey sha _) = do
+      mbid <- loadBlobBySHA sha
+      case mbid of
+        Nothing -> throwIO $ MissingBlob blobKey
+        Just bid -> pure bid
