@@ -1,4 +1,5 @@
 {-# LANGUAGE BangPatterns #-}
+{-# LANGUAGE CPP #-}
 {-# LANGUAGE DeriveFunctor #-}
 {-# LANGUAGE DeriveGeneric #-}
 {-# LANGUAGE FlexibleInstances #-}
@@ -11,6 +12,7 @@
 {-# LANGUAGE ScopedTypeVariables #-}
 {-# LANGUAGE TupleSections #-}
 {-# LANGUAGE MultiWayIf #-}
+{-# LANGUAGE NamedFieldPuns #-}
 module Pantry.Types
   ( PantryConfig (..)
   , HackageSecurityConfig (..)
@@ -60,6 +62,11 @@ module Pantry.Types
   , Archive (..)
   , toRawArchive
   , Repo (..)
+  , AggregateRepo (..)
+  , SimpleRepo (..)
+  , toAggregateRepos
+  , rToSimpleRepo
+  , arToSimpleRepo
   , RepoType (..)
   , parsePackageIdentifier
   , parsePackageName
@@ -86,8 +93,10 @@ module Pantry.Types
   , HpackExecutable (..)
   , WantedCompiler (..)
   --, resolveSnapshotLocation
-  , ltsSnapshotLocation
-  , nightlySnapshotLocation
+  , snapshotLocation
+  , defaultSnapshotLocation
+  , SnapName (..)
+  , parseSnapName
   , RawSnapshotLocation (..)
   , SnapshotLocation (..)
   , toRawSL
@@ -109,6 +118,7 @@ module Pantry.Types
   , getGlobalHintsFile
   , bsToBlobKey
   , warnMissingCabalFile
+  , connRDBMS
   ) where
 
 import RIO
@@ -116,14 +126,12 @@ import qualified Data.Conduit.Tar as Tar
 import qualified RIO.Text as T
 import qualified RIO.ByteString as B
 import qualified RIO.ByteString.Lazy as BL
-import RIO.List (intersperse)
-import RIO.Time (toGregorian, Day, fromGregorianValid, UTCTime)
+import RIO.List (intersperse, groupBy)
+import RIO.Time (toGregorian, Day, UTCTime)
 import qualified RIO.Map as Map
-import qualified RIO.HashMap as HM
 import qualified Data.Map.Strict as Map (mapKeysMonotonic)
 import qualified RIO.Set as Set
-import Data.Aeson (ToJSON (..), FromJSON (..), withText, FromJSONKey (..))
-import Data.Aeson.Types (ToJSONKey (..) ,toJSONKeyText, Parser)
+import Data.Aeson.Types (toJSONKeyText, Parser)
 import Pantry.Internal.AesonExtended
 import Data.Aeson.Encoding.Internal (unsafeToEncoding)
 import Data.ByteString.Builder (toLazyByteString, byteString, wordDec)
@@ -132,7 +140,12 @@ import Database.Persist.Sql
 import Pantry.SHA256 (SHA256)
 import qualified Pantry.SHA256 as SHA256
 import qualified Distribution.Compat.CharParsing as Parse
-import Distribution.CabalSpecVersion (CabalSpecVersion (..), cabalSpecLatest)
+import Distribution.CabalSpecVersion (cabalSpecLatest)
+#if MIN_VERSION_Cabal(3,4,0)
+import Distribution.CabalSpecVersion (cabalSpecToVersionDigits)
+#else
+import Distribution.CabalSpecVersion (CabalSpecVersion (..))
+#endif
 import Distribution.Parsec (PError (..), PWarning (..), showPos, parsec, explicitEitherParsec, ParsecParser)
 import Distribution.Types.PackageName (PackageName, unPackageName, mkPackageName)
 import Distribution.Types.VersionRange (VersionRange)
@@ -148,10 +161,24 @@ import Network.HTTP.Types (Status, statusCode)
 import Data.Text.Read (decimal)
 import Path (Path, Abs, Dir, File, toFilePath, filename, (</>), parseRelFile)
 import Path.IO (resolveFile, resolveDir)
-import Data.List.NonEmpty (NonEmpty)
 import qualified Data.List.NonEmpty as NE
 import Casa.Client (CasaRepoPrefix)
 
+#if MIN_VERSION_persistent(2, 13, 0)
+import Database.Persist.SqlBackend.Internal (connRDBMS)
+#endif
+
+#if MIN_VERSION_aeson(2, 0, 0)
+import qualified Data.Aeson.KeyMap as HM
+import qualified Data.Aeson.Key
+
+type AesonKey = Data.Aeson.Key.Key
+#else
+import qualified RIO.HashMap as HM
+
+type AesonKey = Text
+#endif
+
 -- | Parsed tree with more information on the Haskell package it contains.
 --
 -- @since 0.1.0.0
@@ -175,18 +202,18 @@ data Package = Package
   --
   -- @since 0.1.0.0
   }
-  deriving (Show, Eq)
+  deriving (Show, Eq, Ord)
 
 data PHpack = PHpack
     {
       phOriginal :: !TreeEntry, -- ^ Original hpack file
       phGenerated :: !TreeEntry, -- ^ Generated Cabal file
       phVersion :: !Version -- ^ Version of Hpack used
-    } deriving (Show, Eq)
+    } deriving (Show, Eq, Ord)
 
 data PackageCabal = PCCabalFile !TreeEntry -- ^ TreeEntry of Cabal file
                   | PCHpack !PHpack
-                  deriving (Show, Eq)
+                  deriving (Show, Eq, Ord)
 
 cabalFileName :: PackageName -> SafeFilePath
 cabalFileName name =
@@ -250,8 +277,18 @@ data PantryConfig = PantryConfig
   -- ^ The pull URL e.g. @https://casa.fpcomplete.com/v1/pull@
   , pcCasaMaxPerRequest :: !Int
   -- ^ Maximum blobs sent per pull request.
+  , pcSnapshotLocation :: SnapName -> RawSnapshotLocation
+  -- ^ The location of snapshot synonyms
   }
 
+-- | Get the location of a snapshot synonym from the 'PantryConfig'.
+--
+-- @since 0.5.0.0
+snapshotLocation :: HasPantryConfig env => SnapName -> RIO env RawSnapshotLocation
+snapshotLocation name = do
+  loc <- view $ pantryConfigL.to pcSnapshotLocation
+  pure $ loc name
+
 -- | Should we print warnings when loading a cabal file?
 --
 -- @since 0.1.0.0
@@ -514,6 +551,51 @@ instance Display Repo where
       then mempty
       else " in subdirectory " <> display subdir)
 
+rToSimpleRepo :: Repo -> SimpleRepo
+rToSimpleRepo Repo {..} = SimpleRepo { sRepoUrl = repoUrl, sRepoCommit = repoCommit, sRepoType = repoType }
+
+data AggregateRepo = AggregateRepo
+  { aRepo :: !SimpleRepo
+  , aRepoSubdirs :: [(Text, RawPackageMetadata)]
+  }
+    deriving (Show, Generic, Eq, Ord, Typeable)
+
+-- | Group input repositories by non-subdir values.
+toAggregateRepos :: [(Repo, RawPackageMetadata)] -> [AggregateRepo]
+toAggregateRepos = mapMaybe toAggregateRepo . groupBy matchRepoExclSubdir
+ where
+  toAggregateRepo :: [(Repo, RawPackageMetadata)] -> Maybe AggregateRepo
+  toAggregateRepo [] = Nothing
+  toAggregateRepo xs@((repo, _):_) =
+    Just $ AggregateRepo (rToSimpleRepo repo) (fmap (first repoSubdir) xs)
+
+  matchRepoExclSubdir x1 x2 =
+    let (Repo url1 commit1 type1 _, _) = x1
+        (Repo url2 commit2 type2 _, _) = x2
+    in  (url1, commit1, type1) == (url2, commit2, type2)
+
+arToSimpleRepo :: AggregateRepo -> SimpleRepo
+arToSimpleRepo AggregateRepo {..} = aRepo
+
+-- | Repository without subdirectory information.
+--
+-- @since 0.5.3
+data SimpleRepo = SimpleRepo
+  { sRepoUrl :: !Text
+  , sRepoCommit :: !Text
+  , sRepoType :: !RepoType
+  }
+    deriving (Show, Generic, Eq, Ord, Typeable)
+
+instance Display SimpleRepo where
+  display (SimpleRepo url commit typ) =
+    (case typ of
+       RepoGit -> "Git"
+       RepoHg -> "Mercurial") <>
+    " repo at " <>
+    display url <>
+    ", commit " <>
+    display commit
 
 -- An unexported newtype wrapper to hang a 'FromJSON' instance off of. Contains
 -- a GitHub user and repo name separated by a forward slash, e.g. "foo/bar".
@@ -586,7 +668,7 @@ instance Show BlobKey where
 instance Display BlobKey where
   display (BlobKey sha size') = display sha <> "," <> display size'
 
-blobKeyPairs :: BlobKey -> [(Text, Value)]
+blobKeyPairs :: BlobKey -> [(AesonKey, Value)]
 blobKeyPairs (BlobKey sha size') =
     [ "sha256" .= sha
     , "size" .= size'
@@ -619,7 +701,7 @@ instance FromJSON PackageNameP where
 instance ToJSONKey PackageNameP where
   toJSONKey =
     ToJSONKeyText
-      (T.pack . packageNameString . unPackageNameP)
+      (fromString . packageNameString . unPackageNameP)
       (unsafeToEncoding . getUtf8Builder . display)
 instance FromJSONKey PackageNameP where
   fromJSONKey = FromJSONKeyText $ PackageNameP . mkPackageName . T.unpack
@@ -828,7 +910,7 @@ data PantryException
   | InvalidTarFileType !ArchiveLocation !FilePath !Tar.FileType
   | UnsupportedTarball !ArchiveLocation !Text
   | NoHackageCryptographicHash !PackageIdentifier
-  | FailedToCloneRepo !Repo
+  | FailedToCloneRepo !SimpleRepo
   | TreeReferencesMissingBlob !RawPackageLocationImmutable !SafeFilePath !BlobKey
   | CompletePackageMetadataMismatch !RawPackageLocationImmutable !PackageMetadata
   | CRC32Mismatch !ArchiveLocation !FilePath !(Mismatch Word32)
@@ -842,6 +924,7 @@ data PantryException
   | DuplicatePackageNames !Utf8Builder ![(PackageName, [RawPackageLocationImmutable])]
   | MigrationFailure !Text !(Path Abs File) !SomeException
   | InvalidTreeFromCasa !BlobKey !ByteString
+  | ParseSnapNameException !Text
 
   deriving Typeable
 instance Exception PantryException where
@@ -1026,6 +1109,7 @@ instance Display PantryException where
     "\nPlease report this on https://github.com/commercialhaskell/stack/issues" <>
     "\nAs a workaround you may delete " <> display desc <> " database in " <>
     fromString (toFilePath fp) <> " triggering its recreation."
+  display (ParseSnapNameException t) = "Invalid snapshot name: " <> display t
 
 data FuzzyResults
   = FRNameNotFound ![PackageName]
@@ -1058,32 +1142,34 @@ orSeparated xs
 commaSeparated :: NonEmpty Utf8Builder -> Utf8Builder
 commaSeparated = fold . NE.intersperse ", "
 
--- You'd really think there'd be a better way to do this in Cabal.
 cabalSpecLatestVersion :: Version
-cabalSpecLatestVersion =
-  case cabalSpecLatest of
-    CabalSpecV1_0 -> error "this cannot happen"
-    CabalSpecV1_2 -> error "this cannot happen"
-    CabalSpecV1_4 -> error "this cannot happen"
-    CabalSpecV1_6 -> error "this cannot happen"
-    CabalSpecV1_8 -> error "this cannot happen"
-    CabalSpecV1_10 -> error "this cannot happen"
-    CabalSpecV1_12 -> error "this cannot happen"
-    CabalSpecV1_18 -> error "this cannot happen"
-    CabalSpecV1_20 -> error "this cannot happen"
-    CabalSpecV1_22 -> error "this cannot happen"
-    CabalSpecV1_24 -> error "this cannot happen"
-    CabalSpecV2_0 -> error "this cannot happen"
-    CabalSpecV2_2 -> error "this cannot happen"
-    CabalSpecV2_4 -> error "this cannot happen"
-    CabalSpecV3_0 -> mkVersion [3, 0]
+cabalSpecLatestVersion = mkVersion $ cabalSpecToVersionDigits cabalSpecLatest
+
+#if !MIN_VERSION_Cabal(3,4,0)
+cabalSpecToVersionDigits :: CabalSpecVersion -> [Int]
+cabalSpecToVersionDigits CabalSpecV3_0   = [3,0]
+cabalSpecToVersionDigits CabalSpecV2_4   = [2,4]
+cabalSpecToVersionDigits CabalSpecV2_2   = [2,2]
+cabalSpecToVersionDigits CabalSpecV2_0   = [2,0]
+cabalSpecToVersionDigits CabalSpecV1_24  = [1,24]
+cabalSpecToVersionDigits CabalSpecV1_22  = [1,22]
+cabalSpecToVersionDigits CabalSpecV1_20  = [1,20]
+cabalSpecToVersionDigits CabalSpecV1_18  = [1,18]
+cabalSpecToVersionDigits CabalSpecV1_12  = [1,12]
+cabalSpecToVersionDigits CabalSpecV1_10  = [1,10]
+cabalSpecToVersionDigits CabalSpecV1_8   = [1,8]
+cabalSpecToVersionDigits CabalSpecV1_6   = [1,6]
+cabalSpecToVersionDigits CabalSpecV1_4   = [1,4]
+cabalSpecToVersionDigits CabalSpecV1_2   = [1,2]
+cabalSpecToVersionDigits CabalSpecV1_0   = [1,0]
+#endif
 
 data BuildFile = BFCabal !SafeFilePath !TreeEntry
                | BFHpack !TreeEntry -- We don't need SafeFilePath for Hpack since it has to be package.yaml file
   deriving (Show, Eq)
 
 data FileType = FTNormal | FTExecutable
-  deriving (Show, Eq, Enum, Bounded)
+  deriving (Show, Eq, Enum, Bounded, Ord)
 instance PersistField FileType where
   toPersistValue FTNormal = PersistInt64 1
   toPersistValue FTExecutable = PersistInt64 2
@@ -1101,7 +1187,7 @@ data TreeEntry = TreeEntry
   { teBlob :: !BlobKey
   , teType :: !FileType
   }
-  deriving (Show, Eq)
+  deriving (Show, Eq, Ord)
 
 newtype SafeFilePath = SafeFilePath Text
   deriving (Show, Eq, Ord, Display)
@@ -1159,7 +1245,7 @@ newtype Tree
   -- In the future, consider allowing more lax parsing
   -- See: https://www.fpcomplete.com/blog/2018/07/pantry-part-2-trees-keys
   -- TreeTarball !PackageTarball
-  deriving (Show, Eq)
+  deriving (Show, Eq, Ord)
 
 renderTree :: Tree -> ByteString
 renderTree = BL.toStrict . toLazyByteString . go
@@ -1538,7 +1624,7 @@ instance ToJSON RawPackageLocationImmutable where
           RepoGit -> "git"
           RepoHg  -> "hg"
 
-rpmToPairs :: RawPackageMetadata -> [(Text, Value)]
+rpmToPairs :: RawPackageMetadata -> [(AesonKey, Value)]
 rpmToPairs (RawPackageMetadata mname mversion mtree) = concat
   [ maybe [] (\name -> ["name" .= CabalString name]) mname
   , maybe [] (\version -> ["version" .= CabalString version]) mversion
@@ -1845,9 +1931,7 @@ instance FromJSON (WithJSONWarnings (Unresolved RawSnapshotLocation)) where
 
 instance Display SnapshotLocation where
   display (SLCompiler compiler) = display compiler
-  display (SLUrl url blob) =
-    fromMaybe (display url) (specialRawSnapshotLocation url) <>
-    " (" <> display blob <> ")"
+  display (SLUrl url blob) = display url <> " (" <> display blob <> ")"
   display (SLFilePath resolved) = display (resolvedRelative resolved)
 
 -- | Parse a 'Text' into an 'Unresolved' 'RawSnapshotLocation'.
@@ -1856,22 +1940,10 @@ instance Display SnapshotLocation where
 parseRawSnapshotLocation :: Text -> Unresolved RawSnapshotLocation
 parseRawSnapshotLocation t0 = fromMaybe (parseRawSnapshotLocationPath t0) $
   (either (const Nothing) (Just . pure . RSLCompiler) (parseWantedCompiler t0)) <|>
-  parseLts <|>
-  parseNightly <|>
+  (pure <$> RSLSynonym <$> parseSnapName t0) <|>
   parseGithub <|>
   parseUrl
   where
-    parseLts = do
-      t1 <- T.stripPrefix "lts-" t0
-      Right (x, t2) <- Just $ decimal t1
-      t3 <- T.stripPrefix "." t2
-      Right (y, "") <- Just $ decimal t3
-      Just $ pure $ ltsSnapshotLocation x y
-    parseNightly = do
-      t1 <- T.stripPrefix "nightly-" t0
-      date <- readMaybe (T.unpack t1)
-      Just $ pure $ nightlySnapshotLocation date
-
     parseGithub = do
       t1 <- T.stripPrefix "github:" t0
       let (user, t2) = T.break (== '/') t1
@@ -1909,29 +1981,75 @@ defUser = "commercialhaskell"
 defRepo :: Text
 defRepo = "stackage-snapshots"
 
--- | Location of an LTS snapshot
+-- | Default location of snapshot synonyms
+-- , i.e. commercialhaskell's GitHub repository.
 --
--- @since 0.1.0.0
-ltsSnapshotLocation
-  :: Int -- ^ major version
-  -> Int -- ^ minor version
+-- @since 0.5.0.0
+defaultSnapshotLocation
+  :: SnapName
   -> RawSnapshotLocation
-ltsSnapshotLocation x y =
+defaultSnapshotLocation (LTS x y) =
   githubSnapshotLocation defUser defRepo $
   utf8BuilderToText $
   "lts/" <> display x <> "/" <> display y <> ".yaml"
-
--- | Location of a Stackage Nightly snapshot
---
--- @since 0.1.0.0
-nightlySnapshotLocation :: Day -> RawSnapshotLocation
-nightlySnapshotLocation date =
+defaultSnapshotLocation (Nightly date) =
   githubSnapshotLocation defUser defRepo $
   utf8BuilderToText $
   "nightly/" <> display year <> "/" <> display month <> "/" <> display day <> ".yaml"
   where
     (year, month, day) = toGregorian date
 
+-- | A snapshot synonym.
+-- It is expanded according to the field 'snapshotLocation'
+-- of a 'PantryConfig'.
+--
+-- @ since 0.5.0.0
+data SnapName
+    -- | LTS Haskell snapshot,
+    -- displayed as @"lts-maj.min"@.
+    --
+    -- @since 0.5.0.0
+    = LTS
+        !Int -- ^ Major version
+        !Int -- ^ Minor version
+    -- | Stackage Nightly snapshot,
+    -- displayed as @"nighly-YYYY-MM-DD"@.
+    --
+    -- @since 0.5.0.0
+    | Nightly !Day
+    deriving (Eq, Ord, Generic)
+
+instance NFData SnapName
+
+instance Display SnapName where
+  display (LTS x y) = "lts-" <> display x <> "." <> display y
+  display (Nightly date) = "nightly-" <> displayShow date
+
+instance Show SnapName where
+  show = T.unpack . utf8BuilderToText . display
+
+instance ToJSON SnapName where
+  toJSON syn = String $ utf8BuilderToText $ display syn
+
+-- | Parse the short representation of a 'SnapName'.
+--
+-- @since 0.5.0.0
+parseSnapName :: MonadThrow m => Text -> m SnapName
+parseSnapName t0 =
+    case lts <|> nightly of
+        Nothing -> throwM $ ParseSnapNameException t0
+        Just sn -> return sn
+  where
+    lts = do
+        t1 <- T.stripPrefix "lts-" t0
+        Right (x, t2) <- Just $ decimal t1
+        t3 <- T.stripPrefix "." t2
+        Right (y, "") <- Just $ decimal t3
+        return $ LTS x y
+    nightly = do
+        t1 <- T.stripPrefix "nightly-" t0
+        Nightly <$> readMaybe (T.unpack t1)
+
 -- | Where to load a snapshot from in raw form
 -- (RSUrl could have a missing BlobKey)
 --
@@ -1951,56 +2069,29 @@ data RawSnapshotLocation
     -- ^ Snapshot at a local file path.
     --
     -- @since 0.1.0.0
+  | RSLSynonym !SnapName
+    -- ^ Snapshot synonym (LTS/Nightly).
+    --
+    -- @since 0.5.0.0
   deriving (Show, Eq, Ord, Generic)
 
 instance NFData RawSnapshotLocation
 
 instance Display RawSnapshotLocation where
   display (RSLCompiler compiler) = display compiler
-  display (RSLUrl url Nothing) = fromMaybe (display url) $ specialRawSnapshotLocation url
-  display (RSLUrl url (Just blob)) =
-    fromMaybe (display url) (specialRawSnapshotLocation url) <>
-    " (" <> display blob <> ")"
+  display (RSLUrl url Nothing) = display url
+  display (RSLUrl url (Just blob)) = display url <> " (" <> display blob <> ")"
   display (RSLFilePath resolved) = display (resolvedRelative resolved)
+  display (RSLSynonym syn) = display syn
 
--- | For nicer display purposes: present a 'RawSnapshotLocation' as a
--- short form like lts-13.13 if possible.
-specialRawSnapshotLocation :: Text -> Maybe Utf8Builder
-specialRawSnapshotLocation url = do
-  t1 <- T.stripPrefix "https://raw.githubusercontent.com/commercialhaskell/stackage-snapshots/master/" url
-  parseLTS t1 <|> parseNightly t1
-  where
-    popInt :: Text -> Maybe (Int, Text)
-    popInt t0 =
-      -- Would be nice if this function did overflow checking for us
-      case decimal t0 of
-        Left _ -> Nothing
-        Right (x, rest) -> (, rest) <$> do
-          if (x :: Integer) > fromIntegral (maxBound :: Int)
-            then Nothing
-            else Just (fromIntegral x)
-
-    parseLTS t1 = do
-      t2 <- T.stripPrefix "lts/" t1
-      (major, t3) <- popInt t2
-      (minor, ".yaml") <- T.stripPrefix "/" t3 >>= popInt
-      Just $ "lts-" <> display major <> "." <> display minor
-    parseNightly t1 = do
-      t2 <- T.stripPrefix "nightly/" t1
-      (year, t3) <- popInt t2
-      (month, t4) <- T.stripPrefix "/" t3 >>= popInt
-      (day, ".yaml") <- T.stripPrefix "/" t4 >>= popInt
-      date <- fromGregorianValid (fromIntegral year) month day
-      Just $ "nightly-" <> displayShow date
 
 instance ToJSON RawSnapshotLocation where
   toJSON (RSLCompiler compiler) = object ["compiler" .= compiler]
-  toJSON (RSLUrl url Nothing)
-    | Just x <- specialRawSnapshotLocation url = String $ utf8BuilderToText x
   toJSON (RSLUrl url mblob) = object
     $ "url" .= url
     : maybe [] blobKeyPairs mblob
   toJSON (RSLFilePath resolved) = object ["filepath" .= resolvedRelative resolved]
+  toJSON (RSLSynonym syn) = toJSON syn
 
 -- | Where to load a snapshot from.
 --
